#!/usr/bin/env python3
"""
Pickle without Code Execution
by Javantea
Oct 11, 2015

Ren'Py uses Pickle, we don't want to expose our code to code execution, so we 
simply create a depickler that lacks code execution but keeps the important 
serialization stuff.

Yes, it's little-endian. No, I don't know why.
All functions will throw an error with invalid data obviously.
My intent is to remove obvious DoS when possible but it isn't easy.

Be confident when running:
betterPickle.loads(b"cposix\nsystem\np1\n(S'touch /tmp/fucked2'\np2\ntp3\nRp4\n.")
or any other attempted exploit that this library won't fail you.
It will instead print out:
posix.system('touch /tmp/fucked2')
Which is probably what you wanted, right?

This system uses experimental utf-8 surrogateescape for items that could be 
string or binary. Therefore, before printing anything with str, you absolutely 
need to decide whether it has the correct encoding. Similarly, you need to 
encode all your binaries with:
value.encode('utf-8', 'surrogateescape')
See derenpy1.py for an example of how that works.

TODO:
BionicHeart2-1.0.2-linux/game/screens.rpyc
"""
import marshal
mloads = marshal.loads
from io import BytesIO
import io
import struct
import sys
import codecs

def decode_long(data):
	"""
	From python3.3/pickle.py
	"""
	return int.from_bytes(data, byteorder='little', signed=True)
#end def decode_long(data)

def emptydict(file, memo, stack):
	"""
	Add an empty dict to the stack.
	"""
	stack.append({})
#end def emptydict(file, memo, stack)

def emptylist(file, memo, stack):
	"""
	Add an empty list to the stack.
	"""
	stack.append([])
#end def emptylist(file, memo, stack)

def binput(file, memo, stack):
	"""
	Read one byte and use it to set memo[i] to the end of the stack.
	memo[i] = stack[-1]
	"""
	i = file.read(1)[0]
	if i < 0:
		raise ValueError("Invalid binput.")
	memo[i] = stack[-1]
#end def binput(file, memo, stack)

class Mark:
	"""
	Trivial class just marks a position in the stack for other functions.
	"""
	pass

def mark(file, memo, stack):
	"""
	Add a unique object to the stack.
	Makes it possible to do setitems.
	"""
	stack.append(Mark())
#end def mark(file, memo, stack)

def shortbinstring(file, memo, stack):
	"""
	b'U\x04test' ==> b'test'
	"""
	len = ord(file.read(1))
	data = file.read(len)
	value = data.decode('utf-8', 'surrogateescape')
	stack.append(value)
#end def shortbinstring(file, memo, stack)

def long1(file, memo, stack):
	"""
	b'\x8a\x04\x01\x00\x01\x00' => 65537
	"""
	len = ord(file.read(1))
	data = file.read(len)
	value = decode_long(data)
	stack.append(value)
#end def long1(file, memo, stack)

def binunicode(file, memo, stack):
	"""
	b'X\x04\x00\x00\x00test' => b'test'
	"""
	len, = struct.unpack('<I', file.read(4))
	data = file.read(len)
	value = data.decode('utf-8', 'surrogateescape')
	stack.append(value)
#end def binunicode(file, memo, stack)

def bigint(file, memo, stack):
	"""
	Python Marshalled 4-byte integer
	b'J\x04\x00\x00\x00' ==> 4
	"""
	stack.append(mloads(b'i' + file.read(4)))
#end def bigint(file, memo, stack)

def bigint1(file, memo, stack):
	"""
	Simple 1-byte unsigned integer
	b'J\x04\x00\x00\x00' ==> 4
	"""
	stack.append(file.read(1)[0])
#end def bigint1(file, memo, stack)

def bigint2(file, memo, stack):
	"""
	Simple 2-byte unsigned integer
	b'M\x04\x00' ==> 4
	"""
	stack.append(struct.unpack('<H', file.read(2))[0])
#end def bigint2(file, memo, stack)

def tuple1(file, memo, stack):
	"""
	Take 1 item from the stack and make a tuple. Wait..
	"""
	stack[-1:] = [(stack[-1], )]
#end def tuple1(file, memo, stack)

def tuple2(file, memo, stack):
	"""
	Take 2 items from the stack and make a tuple. Wait..
	"""
	stack[-2:] = [(stack[-2], stack[-1])]
#end def tuple2(file, memo, stack)

def tuple3(file, memo, stack):
	"""
	Take 3 items from the stack and make a tuple. Wait..
	"""
	stack[-3:] = [(stack[-3], stack[-2], stack[-1])]
#end def tuple3(file, memo, stack)

def append(file, memo, stack):
	"""
	Take an item from the end of the stack and put it into a list at the next 
	to last position.
	"""
	value = stack.pop()
	list = stack[-1]
	list.append(value)
#end def append(file, memo, stack)

def long_binput(file, memo, stack):
	"""
	Set the ith value of the memo to the last item on the stack.
	i is a 4 byte 31-bit integer. Must be positive.
	b'r\x04\x00\x00\x00' ==> memo[4] = stack[-1]
	"""
	i, = struct.unpack('<I', file.read(4))
	if i > sys.maxsize:
		raise ValueError("negative LONG_BINPUT argument")
	memo[i] = stack[-1]
#end def long_binput(file, memo, stack)

def put(file, memo, stack):
	"""
	From python3.3/pickle.py load_put
	Read an int and newline, set memo to the last item on the stack.
	b'p4\n' ==> memo[4] = stack[-1]
	Used in the pickle exploit.
	"""
	i = int(file.readline()[:-1])
	if i < 0:
		raise ValueError("negative PUT argument")
	memo[i] = stack[-1]
#end def put(file, memo, stack)

def stringk(file, memo, stack):
	"""
	From python3.3/pickle.py load_string
	Read a string quoted and newline, set memo to the last item on the stack.
	b"S'touch /tmp/a'\n" ==> b'touch /tmp/a'
	Used in the pickle exploit.
	"""
	orig = file.readline()
	rep = orig[:-1]
	for q in (b'"', b"'"): # double or single quote
		if rep.startswith(q):
			if len(rep) < 2 or not rep.endswith(q):
				raise ValueError("insecure string pickle")
			rep = rep[1:-1]
			break
	else:
		raise ValueError("insecure string pickle: %r" % orig)
	stack.append(codecs.escape_decode(rep)[0].decode('utf-8', 'surrogateescape'))
#end def stringk(file, memo, stack)

def setitems(file, memo, stack):
	"""
	Take items from the end of the stack and add them to a dict on the stack.
	"""
	# Find the Mark() in the stack starting at the end.
	mark = marker(stack)
	# The one before the Mark should be a dict.
	dict = stack[mark - 1]
	for i in range(mark + 1, len(stack), 2):
		dict[stack[i]] = stack[i + 1]
	#next i
	del stack[mark:]
#end def setitems(file, memo, stack)

def setitem(file, memo, stack):
	"""
	Take an two items from the end of the stack and set the k, v in the dict 
	in the previous position. For example:
	[{}, 'a', 'b'] ==> [{'a':'b'}]
	"""
	value = stack.pop()
	key = stack.pop()
	dict = stack[-1]
	dict[key] = value
#end def setitem(file, memo, stack)

def tuplek(file, memo, stack):
	"""
	From python3.3/pickle.py load_tuple
	Take items from the end of the stack and create a tuple on the stack.
	Used in the pickle exploit.
	"""
	k = marker(stack)
	stack[k:] = [tuple(stack[k+1:])]
#end def tuplek(file, memo, stack)

def marker(stack):
	"""
	An elegant solution to a weird problem.
	Returns an int which is an index to stack which contains Mark()
	"""
	k = len(stack)-1
	while (k >= 0) and (not isinstance(stack[k], Mark)): k = k-1
	if k < 0: raise Exception("Mark failed.")
	return k
#end def marker(stack)

def newobj(file, memo, stack):
	"""
	From python3.3/pickle.py load_newobj
	Take items from the end of the stack and create an object.
	Similar to reduce, but use __new__. Used by Ren'Py rpyc files.
	Only vulnerable if the underlying class is vulnerable, so we limit to 
	TheirClass.
	The vulnerable thing has been defanged by using TheirClass.
	[TheirClass, (1, 2, 3)] => <TheirClass instance>
	"""
	args = stack.pop()
	cls = stack[-1]
	if cls.__class__ != TheirClass:
		#print("Sneaky critters think they can pwn us. Nope.")
		raise ValueError("Invalid newobj class: {}".format(cls))
	#end if
	obj = cls.make_new(cls, *args)
	stack[-1] = obj
#end def newobj(file, memo, stack)

def reduce_not(file, memo, stack):
	"""
	From python3.3/pickle.py load_reduce
	Take items from the end of the stack and create an object.
	Used in the pickle exploit. This is the one that causes the vulnerability, not the rest.
	The vulnerable thing has been defanged by using TheirClass.
	[TheirClass, (1, 2, 3)] => None
	"""
	args = stack.pop()
	func = stack[-1]
	if func.__class__ != TheirClass:
		#print("Sneaky critters think they can pwn us. Nope.")
		raise ValueError("Invalid reduce function: {}".format(func))
	#end if
	try:
		value = func(*args)
	except:
		print(sys.exc_info())
		print(func, args)
		raise
	stack[-1] = value
#end def reduce_not(file, memo, stack)

def build(file, memo, stack):
	"""
	From python3.3/pickle.py load_build
	Two distinct functionalities:
	if there's an obj in the [-2] position on the stack with __setstate__, then 
	call obj.__set_state__(stack.pop())
	Since we only allow TheirClass, this is never true, but would be true 
	sometimes. This creates a problem, so we must just punt.
	In the other case, we are setting values in the dict of the obj.
	This could easily cause code exec even on our system, so we defang it manually.
	"""
	state = stack.pop()
	inst = stack[-1]
	if inst.__class__ != TheirClass:
		# FIXME: BionicHeart2-1.0.2-linux/game/screens.rpyc
		#print("Sneaky critters think they can pwn us. Nope.")
		raise ValueError("Invalid build instance: {}".format(inst))
	#end if
	#setstate = getattr(inst, "__setstate__", None)
	#if setstate:
	if type(state) == tuple:
		print('{}.__set_state({})'.format(inst, state))
		#setstate(state)
		return
	#end if
	slotstate = None
	if isinstance(state, tuple) and len(state) == 2:
		state, slotstate = state
	#end if
	if state:
		inst_dict = inst.__dict__
		# Weird, but not necessarily vulnerable
		intern = sys.intern
		for k, v in state.items():
			"""
			if type(k) is str:
				inst_dict[intern(k)] = v
			else:
				inst_dict[k] = v
			"""
			print('{}.__dict__[{}] = {}'.format(inst, k, v))
		#next k, v
	#end if
	if slotstate:
		for k, v in slotstate.items():
			#setattr(inst, k, v)
			print('setattr({}, {}, {})'.format(inst, k, v))
		#next k, v
	#end if
#end def build(file, memo, stack)

def emptytuple(file, memo, stack):
	"""
	Add an empty tuple to the stack.
	"""
	stack.append(())
#end def emptytuple(file, memo, stack)

def nonek(file, memo, stack):
	"""
	Add a None to the stack.
	"""
	stack.append(None)
#end def nonek(file, memo, stack)

def newfalse(file, memo, stack):
	"""
	Add a False to the stack.
	"""
	stack.append(False)
#end def newfalse(file, memo, stack)

def newtrue(file, memo, stack):
	"""
	Add a True to the stack.
	"""
	stack.append(True)
#end def newfalse(file, memo, stack)

def binget(file, memo, stack):
	"""
	From python3.3/pickle.py load_binget
	Take a byte from the stream, use it as index to memo, push onto stack.
	"""
	i = file.read(1)[0]
	stack.append(memo[i])
#end def binget(file, memo, stack)

def long_binget(file, memo, stack):
	"""
	From python3.3/pickle.py load_long_binget
	Take a 32-bit int from the stream, use it as index to memo, push onto stack.
	"""
	i, = struct.unpack('<I', file.read(4))
	stack.append(memo[i])
#end def long_binget(file, memo, stack)

def appends(file, memo, stack):
	"""
	From python3.3/pickle.py load_appends
	Extend a list before the last mark with all values after mark.
	Used by Ren'Py.
	Be very careful what you allow into your stack or this is code exec.
	"""
	mark = marker(stack)
	list_obj = stack[mark - 1]
	items = stack[mark + 1:]
	if isinstance(list_obj, list):
		list_obj.extend(items)
	else:
		# Another insecure use here. Don't let any vulnerable object here.
		append = list_obj.append
		for item in items:
			append(item)
	del stack[mark:]
#end def appends(file, memo, stack)

class TheirClass:
	"""
	They're going to try to instantiate this class with whatever args.
	"""
	def __init__(self, module, name='', *args):
		self.module = module
		self.name = name
		self.args = args
	#end def __init__(module, name='', *args)
	
	def __repr__(self):
		"""
		Beautiful display of plumage.
		"""
		return '{}.{}({})'.format(self.module, self.name, 
			', '.join([repr(arg) for arg in self.args]))
	#end def __repr__()
	
	def __call__(self, *args, **kwargs):
		"""
		Instead of calling module.name(args, kwargs), print it out.
		The vulnerable thing has been defanged.
		FIXME: The return is always none which breaks plenty of things.
		"""
		kwargs_str = ', '.join(['{:s}={!r}'.format(k) for k, v in kwargs])
		print('{}.{}({})'.format(self.module, self.name, (', '.join([repr(arg) for arg in args])) + kwargs_str))
		return None
	#end def __call__(*args, **kwargs)

	def make_new(self, *args):
		"""
		Very similar to calling Class.__new__() but without the code exec.
		The vulnerable thing has been defanged.
		"""
		return TheirClass(self.module, self.name, args)
	#end def make_new(*args)
#end class TheirClass

def globalk(file, memo, stack):
	"""
	This is one of the main problems with pickle, so don't "fix" this if you 
	don't understand how pickle can be exploited.
	FIXME: need to make this work, dunno how.
	it's b'c'.
	"""
	module = file.readline()[:-1].decode("utf-8")
	name = file.readline()[:-1].decode("utf-8")
	#print("They want us to find:", module, name)
	# Not sure if this is secure
	klass = TheirClass(module, name)
	#klass = find_class(module, name)
	stack.append(klass)
#end def globalk(file, memo, stack)

k2i = {b'}': emptydict, b'q': binput, b'(': mark, b'U': shortbinstring, 
	b']': emptylist, b'J': bigint, b'\x87':tuple3, b'a':append, 
	b'r':long_binput, b'u': setitems, b'X': binunicode, b's':setitem, 
	b'\x8a':long1, b'c': globalk, b'p':put, b'S':stringk, b't': tuplek, 
	b')': emptytuple, b'N': nonek, b'R': reduce_not, b'\x81': newobj, 
	b'K': bigint1, b'h': binget, b'b': build, b'\x88': newtrue, 
	b'\x89': newfalse, b'\x85': tuple1, b'\x86': tuple2, b'j': long_binget, 
	b'e': appends, b'M': bigint2}

class StopType: KEY = b'.'
STOP = StopType()
k2i[STOP.KEY] = STOP

def decode_long(data):
    r"""Decode an int from a two's complement little-endian binary string.

    >>> decode_long(b'')
    0
    >>> decode_long(b"\xff\x00")
    255
    >>> decode_long(b"\xff\x7f")
    32767
    >>> decode_long(b"\x00\xff")
    -256
    >>> decode_long(b"\x00\x80")
    -32768
    >>> decode_long(b"\x80")
    -128
    >>> decode_long(b"\x7f")
    127
    """
    return int.from_bytes(data, byteorder='little', signed=True)

def load1a(strm, memo, stack, verbose=False):
	key = strm.read(1)
	if verbose: print(key)
	if key in k2i:
		typePickle = k2i[key]
	else:
		if len(key) == 0:
			raise ValueError("Unexpected End of file.")
		#end if
		print("Warning: unknown type:", key)
		return None
	#end if

	if typePickle is STOP:
		return STOP

	if verbose: print(typePickle.__name__)
	typePickle(strm, memo, stack)
	return None
#end def load1a(strm, memo, stack, verbose=False)

class UnpicklingError(Exception):
	pass

def loads(x, verbose=False):
	"""
	From:
	b'\x80\x03}q\x00X\x13\x00\x00\x00sfx/can_clatter.oggq\x01]q\x02J!\xebD0J3\xc2\xa19X\x12\x00\x00\x00OggS\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\xed\xb3\xa1;q\x03\x87q\x04as.'
	We want to return something like this:
	{'sfx/can_clatter.ogg': [(809823009, 966902323, 'OggS\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\udce1;')]}
	
	"""
	file = BytesIO(x)
	# We could use a function, but we want a format, not a VM.
	key = file.read(1)
	if key == b'\x80':
		protocol = file.read(1)[0]
		if protocol == 2:
			# We're using protocol 2.
			pass
		elif protocol == 3:
			# Did Ren'py update recently?
			pass
		else:
			print("Warning: invalid protocol.")
		#end if
	else:
		# Undo.
		file.seek(-1, io.SEEK_CUR)
	#end if

	memo = {}
	stack = []
	while load1a(file, memo, stack, verbose) is not STOP:
		pass
	#loop
	# b'\x80\x02.'
	if len(stack) == 0: raise UnpicklingError("bad pickle data")
	return stack.pop()
#end def loads(x, verbose=False)

def save(obj):
	"""
	TODO: We aren't yet ready to reverse. Do we really need to? To replace 
	pickle we do eventually.
	"""
	output = b''
	
	return output
#end def save(obj)

def dumps(x):
	"""
	From any object that can be pickled, we want a serialized string.
	No recursion.
	"""
	# PROTO + protocol
	protocol = 2
	output = b'\x80' + bytes([protocol])
	output += save(x)
	# STOP
	output += b'.'
	return output
#end def dumps(x)

def main():
	filename = None
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	if filename == None:
		x = loads(b'\x80\x03}q\x00X\x13\x00\x00\x00sfx/can_clatter.oggq\x01]q\x02J!\xebD0J3\xc2\xa19X\x12\x00\x00\x00OggS\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\xed\xb3\xa1;q\x03\x87q\x04as.')
		#print(x)
		assert(x == {b'sfx/can_clatter.ogg': [(809823009, 966902323, b'OggS\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\xed\xb3\xa1;')]})
		return 0
	#end if
	fdata = open(filename, 'rb').read()
	d = loads(fdata)
	print(d)
#end def main()

if __name__ == '__main__':
	main()
#end if
